#    Copyright (C) 2005 Paul Harrison
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""\
This is a package for calculation of Levy stable distributions and for
fitting these distributions to data.

It operates by interpolating values from a table, as direct computation 
of these distributions requires a lengthy numerical integration. This 
interpolation scheme allows fast fitting of Levy stable distributions 
to data using the Maximum Likelihood technique.

Does not support alpha values less than 0.5.
"""


__version__ = "0.1"

try:
    from numarray import *
except:
    from Numeric import *


def _make_range(min, max, n):
    """ Create a sequence of values. """
    return (arange(n)) * ((max-min)/(n-1.0)) + min


def _calculate_levy(x, alpha, beta):
    """ Calculation of Levy stable distribution via numerical integration.
        This is used in the creation of the lookup table. """
    # "0" parameterization as per http://academic2.american.edu/~jpnolan/stable/stable.html
    # Note: fails for alpha=1.0
    #       (so make sure alpha=1.0 isn't exactly on the interpolation grid)
    from scipy import integrate

    C = beta * tan(pi*0.5*alpha)
    
    def func_cos(u):
        ua = u ** alpha
        return exp(-ua)*cos(C*ua-C*u)

    def func_sin(u):
        ua = u ** alpha
        return exp(-ua)*sin(C*ua-C*u)

    return ( integrate.quad(func_cos, 0.0, integrate.Inf, weight="cos", wvar=x, limlst=1000)[0]
           - integrate.quad(func_sin, 0.0, integrate.Inf, weight="sin", wvar=x, limlst=1000)[0] ) / pi


def _levy_tan(x, alpha, beta):
    """ Calculate the values stored in the lookup table. 
        The tan mapping allows the table to cover the range from -INF to INF. """
    x = tan(x)
    return _calculate_levy(x,alpha,beta)


def _interpolate(points, grid, lower, upper):
    """ Perform multi-dimensional Catmull-Rom cubic interpolation. """
    grid_shape = array(shape(grid))
    dims = len(grid_shape)
    
    points = (points-lower) * ((grid_shape-1) / (upper-lower))
    
    floors = asarray(floor(points),Int)
    
    offsets = points - floors
    offsets2 = offsets * offsets
    offsets3 = offsets2 * offsets
    weighters = [
        -0.5*offsets3    +offsets2-0.5*offsets,
         1.5*offsets3-2.5*offsets2            +1.0, 
        -1.5*offsets3  +2*offsets2+0.5*offsets,
         0.5*offsets3-0.5*offsets2,
    ]
    
    ravel_grid = ravel(grid)

    result = zeros(shape(points)[:-1], Float)
    for i in xrange(1 << (dims*2)):
        weights = ones(shape(points)[:-1], Float)
        ravel_offset = 0
        for j in xrange(dims):
            n = (i >> (j*2)) % 4
            ravel_offset = ravel_offset * grid_shape[j] + \
                maximum(0,minimum(grid_shape[j]-1,floors[:,j] + (n-1)))
            weights *= weighters[n][...,j]
            
        result += weights * take(ravel_grid, ravel_offset)

    return result



# Dimensions: 0 - x, 1 - alpha, 2 - beta
_lower = array([-pi/2 * 0.999, 0.5, -1.0])
_upper = array([pi/2 * 0.999, 2.0, 1.0])

def _make_data_file():
    """ Generates the lookup table, writes it to a .py file. """
    size = (200, 50, 51)
    matrix = zeros(size, Float)
    ranges = [ _make_range(_lower[i],_upper[i],size[i]) for i in xrange(len(size)) ]

    print "Generating levy_data.py ..."
    for i in xrange(size[1]):
        for j in xrange(size[2]):
            print "Calculating alpha =", ranges[1][i], "beta = ", ranges[2][j]
            for k in xrange(size[0]):
                matrix[k,i,j] = _levy_tan(ranges[0][k], ranges[1][i], ranges[2][j])

    file = open("levy_data.py", "wt")
    file.write("""
# This is a generated file, do not edit.
try:
    from numarray import array
except:
    from Numeric import array

matrix = array(%s)\n""" % repr(matrix.tolist()))
    file.close()




def levy(x, alpha, beta):
    """ Interpolate densities of the Levy stable distribution specified by alpha and beta. 
    
        Note: may sometimes return slightly negative values, due to numerical inaccuracies.
    """
    import levy_data

    points = zeros(shape(x) + (3,), Float)
    points[..., 0] = arctan(x) 
    points[..., 1] = alpha
    points[..., 2] = beta
    
    return _interpolate(points, levy_data.matrix, _lower, _upper)


def neglog_levy(x, alpha, beta):
    """ Interpolate negative log densities of the Levy stable distribution specified by alpha and beta. 
    
        Small/negative densities are capped at 1e-100 to preserve sanity.
    """
    return -log(maximum(1e-100, levy(x, alpha, beta)))


def _reflect(x, lower, upper):
    while 1:
        if x < lower:
            x = lower - (x-lower)
        elif x > upper:
            x = upper - (x-upper)
        else:
            return x


def fit_levy(x, alpha=None, beta=None, location=None, scale=None):
    """ Estimate parameters of Levy stable distribution given data x,
        using the Maximum Likelihood method. 
        
        By default, searches all possible Levy stable distributions. However 
        you may restrict the search by specifying the values of one or more
        parameters.
        
        Examples:
        
            levy(x) -- Fit a stable distribution to x
            
            levy(x, beta=0.0) -- Fit a symmetric stable distribution to x
            
            levy(x, beta=0.0, location=0.0) -- Fit a symmetric distribution centered on zero to x
            
            levy(x, alpha=1.0, beta=0.0) -- Fit a Cauchy distribution to x
        
        Returns a tuple of (alpha, beta, location, scale, negative log density)        
        """

    from scipy import optimize
    
    if location == None or scale == None:
        x = sort(x)
        
        last = len(x)-1
        guess_location = x[last/2]
        guess_scale = (x[last-last/4] - x[last/4])/2.0
        
    parameters = [ ]
    
    if alpha != None:
        get_alpha = lambda parameters: alpha
    else:
        get_alpha = lambda parameters, nth=len(parameters): \
            _reflect(parameters[nth],_lower[1],_upper[1])
        parameters.append(1.0)

    if beta != None:
        get_beta = lambda parameters: beta
    else:
        get_beta = lambda parameters, nth=len(parameters): \
            _reflect(parameters[nth],_lower[2],_upper[2])
        parameters.append(0.0)
        
    if location != None:
        get_location = lambda parameters: location
    else:
        get_location = lambda parameters, nth=len(parameters): parameters[nth]
        parameters.append(guess_location)
    
    if scale != None:
        get_scale = lambda parameters: scale
    else:
        get_scale = lambda parameters, nth=len(parameters): exp(parameters[nth])
        parameters.append(log(guess_scale))
        
    def neglog_density(parameters):
        location = get_location(parameters)
        scale = get_scale(parameters)
        return sum(neglog_levy((x-location)/scale, get_alpha(parameters), get_beta(parameters))) \
               + (len(x) * log(scale))
              
    parameters = optimize.fmin(neglog_density, parameters, disp=0)

    return (get_alpha(parameters), get_beta(parameters), 
            get_location(parameters), get_scale(parameters),
            neglog_density(parameters))


if __name__ == "__main__":
    if "build" in sys.argv[1:]:
        _make_data_file()
        
    print "Testing fit_levy:"
        
    print fit_levy(array([1.0,2.0,3.0,4.0,5.0,6.0,7.0]))
    
    try:
        import pylab
        for alpha in _make_range(0.5, 2.0, 15):
            for beta in _make_range(-1.0, 1.0, 15):
                pylab.plot(levy(_make_range(-5.0,5.0,1000), alpha, beta), 'g')
            
        pylab.show()
    except:
        print "No matplotlib, no pretty graph."
